@@ -46,7 +46,7 @@ no warnings ;
   
   my $THI_ID : shared ;
   
-  use vars qw($MOTHER_THREAD %THI_SHARE_TABLE %THI_THREAD_TABLE) ;
+  use vars qw($MOTHER_THREAD %THI_SHARE_TABLE %THI_THREAD_TABLE %GLOBAL_ATTRS) ;
 
   share($MOTHER_THREAD) ;
   
@@ -55,6 +55,8 @@ no warnings ;
   $THI_SHARE_TABLE{tid} = share_new_ref('%') ;
   $THI_SHARE_TABLE{thread} = share_new_ref('%') ;
   $THI_SHARE_TABLE{creator} = share_new_ref('%') ;
+  
+  share(%GLOBAL_ATTRS) ;
 
 #######################
 # START_MOTHER_THREAD #
@@ -108,11 +110,15 @@ sub new {
   $this = bless(share_new_ref('%') , $class) ;
   
   $this->{jobs} = share_new_ref('@') ;
+  $this->{jobs_sz} = share_new_ref('$') ;
+  $this->{jobs_standby} = share_new_ref('@') ;
   $this->{job_id} = share_new_ref('$') ;
   $this->{job_now} = share_new_ref('$') ;
   $this->{status} = share_new_ref('$') ;
   $this->{err} = share_new_ref('$') ;
   
+  $this->{attrs} = share_new_ref('%') ;
+  
   $this->{id} = ++$THI_ID ;
   
   my $shares = share_new_ref('@') ;
@@ -305,6 +311,109 @@ sub id {
   return $this->{id} ;
 }
 
+############
+# SET_ATTR #
+############
+
+sub set_attr {
+  my $this = shift ;
+  my $key = shift ;
+  my $val = shift ;
+
+  share_ref_tree($val) if ref $val ;
+  
+  lock( %{ $this->{attrs} } ) ;
+  
+  return $this->{attrs}{$key} = $val ;
+}
+
+############
+# GET_ATTR #
+############
+
+sub get_attr {
+  my $this = shift ;
+  my $key = shift ;
+  
+  lock( %{ $this->{attrs} } ) ;
+  
+  return $this->{attrs}{$key} ;
+}
+
+##############
+# SET_GLOBAL #
+##############
+
+sub set_global {
+  my $this = shift ;
+  my $key = shift ;
+  my $val = shift ;
+
+  share_ref_tree($val) if ref $val ;
+  
+  lock( %GLOBAL_ATTRS ) ;
+  return $GLOBAL_ATTRS{$key} = $val ;
+}
+
+##############
+# GET_GLOBAL #
+##############
+
+sub get_global {
+  my $this = shift ;
+  my $key = shift ;
+  
+  lock( %GLOBAL_ATTRS ) ;
+  return $GLOBAL_ATTRS{$key} ;
+}
+
+##################
+# SHARE_REF_TREE #
+##################
+
+sub share_ref_tree {
+  my $ref = shift ;
+  
+  my $ref_type = ref $ref ;
+  
+  if ( $ref_type !~ /^(?:ARRAY|HASH|SCALAR|)$/) {
+    if    ( UNIVERSAL::isa($ref , 'ARRAY') )  { $ref_type = 'ARRAY' ;}
+    elsif ( UNIVERSAL::isa($ref , 'HASH') )   { $ref_type = 'HASH' ;}
+    elsif ( UNIVERSAL::isa($ref , 'SCALAR') ) { $ref_type = 'SCALAR' ;}
+    else { return ;}
+  }
+  
+  if ( $ref_type eq 'ARRAY' ) {
+    {
+      eval { lock( @$ref ) } ;
+      share(@$ref) if $@ ;
+    }
+    foreach my $ref_i ( @$ref ) {
+      share_ref_tree($ref_i) if ref $ref_i ;
+    }
+  }
+  elsif ( $ref_type eq 'HASH' ) {
+    {
+      eval { lock( %$ref ) } ;
+      share(%$ref) if $@ ;
+    }
+    foreach my $Key ( keys %$ref ) {
+      share_ref_tree( $$ref{$Key} ) if ref $$ref{$Key} ;
+    }
+  }
+  elsif ( $ref_type eq 'SCALAR' ) {
+    {
+      eval { lock( $$ref ) } ;
+      share($$ref) if $@ ;
+    }
+    share_ref_tree( $$ref ) if ref $$ref ;
+  }
+  
+  $@ = undef ;
+  
+  return $ref ;
+}
+
 ######################
 # IS_RUNNING_ANY_JOB #
 ######################
@@ -313,17 +422,39 @@ sub is_running_any_job {
   my $this = shift ;
   
   return if !$this->exists ;
-  
+    
   my ($jobs , $job_now) = @$this{qw(jobs job_now)} ;
-  
+    
   {
     lock( $$job_now ) ;
     return 1 if defined $$job_now ;
   }
   
+  ## Creates deadlock!!!
+  # {
+  #   lock( @$jobs ) ;
+  #   return 1 if join('',@$jobs) ;
+  # }
+  
+  return 1 if ${$this->{jobs_sz}} ;
+    
+  return ;
+}
+
+###################
+# HAS_JOB_WAITING #
+###################
+
+sub has_job_waiting {
+  my $this = shift ;
+  
+  return if !$this->exists ;
+  
+  my ($jobs) = @$this{qw(jobs)} ;
+  
   {
     lock( @$jobs ) ;
-    return 1 if @$jobs ;
+    return 1 if ${$this->{jobs_sz}} ;
   }
   
   return ;
@@ -406,7 +537,7 @@ sub add_job {
     $the_job = Thread::Isolate::Job->new( $this , $job_type , @_ ) ;
     
     ##print "ADD>> ". $the_job->dump ."\n" ;
-  
+    
     lock( @$jobs ) ;
     
     ##push(@$jobs , $the_job) ;
@@ -418,6 +549,34 @@ sub add_job {
   return $the_job ;
 }
 
+###################
+# ADD_STANDBY_JOB #
+###################
+
+sub add_standby_job {
+  my $this = shift ;
+  my $job_type = shift ;
+  
+  return if !$this->exists ;
+  
+  my ($jobs_standby) = @$this{qw(jobs_standby)} ;
+  
+  my $the_job ;
+  
+  {
+    my $wantarray = shift(@_) ;
+    my $delay = $_[0] =~ /^\d+$/s? shift(@_) : '*' ;
+        
+    $the_job = Thread::Isolate::Job->new( $this , $job_type , $wantarray , @_ ) ;
+    $the_job->detach($delay) ;
+    
+    lock( @$jobs_standby ) ;
+    push(@$jobs_standby , $the_job) ;
+  }
+  
+  return $the_job ;
+}
+
 ###################
 # ADD_JOB_NO_LOCK #
 ###################
@@ -447,9 +606,8 @@ sub _jobs_shift {
   my $this = shift ;
   my ($jobs) = @$this{qw(jobs)} ;
 
+  my $job ;
   { lock( @$jobs ) ;
-  
-    my $job ;
 
     my $i = -1 ;
     $job = $$jobs[++$i] while !$job && $i <= $#{$jobs} ;
@@ -461,9 +619,13 @@ sub _jobs_shift {
       
       @$jobs = map { ($_ ? $_ : ()) } @jobs if @jobs ;
     }
-    
-    return $job ;
+
+    --${$this->{jobs_sz}} if $job ;
+
+    ##print "SHIFT>> ${$this->{jobs_sz}} [". join('',@$jobs) ."]\n" ;
   }
+  
+  return $job ;
 }
 
 ##############
@@ -474,10 +636,16 @@ sub _jobs_push {
   my $this = shift ;
   my ( $the_job ) = @_ ;
   
+  return if !$the_job ;
+  
   my ($jobs) = @$this{qw(jobs)} ;
 
   { lock( @$jobs ) ;
     $$jobs[ $#{$jobs}+1 ] = $the_job ;
+
+    ++${$this->{jobs_sz}} ;
+    
+    #print "PUSH>> ${$this->{jobs_sz}} [". join('',@$jobs) ."]\n" ;
   }
   
   return ;
@@ -490,19 +658,23 @@ sub _jobs_push {
 sub wait_job_to_start {
   my $this = shift ;
   my ( $the_job ) = @_ ;
-
+  
   return if !UNIVERSAL::isa($the_job , 'Thread::Isolate::Job') ;
   return if $this->tid == threads->self->tid || !$this->exists ;
   
-  {
-    lock( @$the_job ) ;
+  { lock( @$the_job ) ;
 
     return 1 if $$the_job[2] >= 1 ;
     
-    cond_wait( @$the_job ) ;
-    #while( !cond_timedwait( @$the_job , time+2 ) ) {
-    #  last if $$the_job[2] >= 1 || ${$this->{status}} <= 0 ;
+    #cond_wait( @$the_job ) ;
+    
+    #while( $$the_job[2] < 1 && ${$this->{status}} > 0 ) {
+    #  select(undef,undef,undef , 0.1);
     #}
+
+    while( !cond_timedwait( @$the_job , time+1 ) ) {
+      last if $$the_job[2] >= 1 || ${$this->{status}} <= 0 ;
+    }
   }
   
   threads->yield while $$the_job[2] < 1 && ${$this->{status}} > 0 ;
@@ -519,17 +691,16 @@ sub wait_job {
 
   return if !UNIVERSAL::isa($the_job , 'Thread::Isolate::Job') ;
   return if $this->tid == threads->self->tid || !$this->exists ;
-  
+
   { lock( @$the_job ) ;
 
     return Thread::Isolate::thaw( $$the_job[4] ) if $$the_job[2] == 2 ;
-
     cond_wait( @$the_job ) ;
     #while( !cond_timedwait( @$the_job , time+2 ) ) {
     #  last if $$the_job[2] == 2 || ${$this->{status}} <= 0 ;
     #}
   }
-  
+    
   threads->yield while $$the_job[2] != 2 && ${$this->{status}} > 0 ;
   return Thread::Isolate::thaw( $$the_job[4] ) ;
 }
@@ -574,7 +745,7 @@ sub call_detached {
 
 sub call {
   my $this = shift ;
-  $this->wait_job( $this->call_detached(@_) ) ;
+  $this->wait_job( ( wantarray ? $this->call_detached(@_) : scalar $this->call_detached(@_) ) ) ;
 }
 
 sub call_detached_no_lock {
@@ -584,9 +755,20 @@ sub call_detached_no_lock {
 
 sub call_no_lock {
   my $this = shift ;
-  $this->wait_job( $this->call_detached_no_lock(@_) ) ;
+  $this->wait_job( ( wantarray ? $this->call_detached_no_lock(@_) : scalar $this->call_detached_no_lock(@_) ) ) ;
 }
 
+sub pack_call_detached {
+  my $this = shift ;
+  @_ = ( _caller_pack() . "::$_[0]" , @_[1..$#_] ) if $_[0] !~ /::/ ;
+  return $this->call_detached(@_) ;
+}
+
+sub pack_call {
+  my $this = shift ;
+  @_ = ( _caller_pack() . "::$_[0]" , @_[1..$#_] ) if $_[0] !~ /::/ ;
+  return $this->call(@_) ;
+}
 
 ########
 # EVAL #
@@ -599,7 +781,7 @@ sub eval_detached {
 
 sub eval {
   my $this = shift ;
-  $this->wait_job( $this->eval_detached(@_) ) ;
+  $this->wait_job( ( wantarray ? $this->eval_detached(@_) : scalar $this->eval_detached(@_) ) ) ;
 }
 
 sub eval_detached_no_lock {
@@ -609,7 +791,60 @@ sub eval_detached_no_lock {
 
 sub eval_no_lock {
   my $this = shift ;
-  $this->wait_job( $this->eval_detached_no_lock(@_) ) ;
+  $this->wait_job( ( wantarray ? $this->eval_detached_no_lock(@_) : scalar $this->eval_detached(@_) ) ) ;
+}
+
+sub pack_eval_detached {
+  my $this = shift ;
+  @_ = ( "package " . _caller_pack() . " ;\n#line1\n$_[0]" , @_[1..$#_] ) if $_[0] !~ /::/ ;
+  return $this->eval_detached(@_) ;
+}
+
+sub pack_eval {
+  my $this = shift ;
+  @_ = ( "package " . _caller_pack() . " ;\n#line1\n$_[0]" , @_[1..$#_] ) if $_[0] !~ /::/ ;
+  return $this->eval(@_) ;
+}
+
+####################
+# ADD_STANDBY_EVAL #
+####################
+
+sub add_standby_eval {
+  my $this = shift ;
+  return $this->add_standby_job('EVAL', (wantarray? 1 : 0) , @_) ;
+}
+
+sub pack_add_standby_eval {
+  my $this = shift ;
+  @_ = ( "package " . _caller_pack() . " ;\n#line1\n$_[0]" , @_[1..$#_] ) if $_[0] !~ /::/ ;
+  return $this->add_standby_eval(@_) ;
+}
+
+####################
+# ADD_STANDBY_CALL #
+####################
+
+sub add_standby_call {
+  my $this = shift ;
+  return $this->add_standby_job('CALL', (wantarray? 1 : 0) , @_) ;
+}
+
+sub pack_add_standby_call {
+  my $this = shift ;
+  @_ = ( _caller_pack() . "::$_[0]" , @_[1..$#_] ) if $_[0] !~ /::/ ;
+  return $this->add_standby_call(@_) ;
+}
+
+################
+# _CALLER_PACK #
+################
+
+sub _caller_pack {
+  my ($i , $pack) = -1 ;
+  $pack = caller(++$i) while $i == -1 || ($pack =~ /^Thread::Isolate(?:::|$)/ && $pack) ;
+  $pack ||= caller || 'main' ;
+  return $pack ;
 }
 
 ############
@@ -660,6 +895,7 @@ sub job_list {
   my ($jobs) = @$this{qw(jobs)} ;
   
   my @list ;
+  
   { lock( @$jobs ) ;
     push(@list , @$jobs) ;
   }
@@ -690,7 +926,7 @@ sub DESTROY {
 ##################
 
 $sub_THREAD_ISOLATE = q`
-#line 694 Thread/Isolate/Thread.pm
+#line 883 Thread/Isolate/Thread.pm
 
 sub {
   my $this = shift ;
@@ -703,32 +939,70 @@ sub {
   
   my $is_mother_thread = $this->{id} == $MOTHER_THREAD ? 1 : undef ;
   
-  my ($jobs , $status) = @$this{qw(jobs status)} ;
+  my ($jobs , $jobs_standby , $status) = @$this{qw(jobs jobs_standby status)} ;
   
+  my $jobs_standby_i = -1 ;
+
   $$status = 1 ;
   
   my $running = 1 ;
+  my $last_job_is_standby ;
   
   while( $running && $$status > 0 ) {
-    lock( @$jobs ) ;
+    my ( $the_job , $standby_job , $standby_job_delay ) ;
+    #print "RUN...>>> $this->{id}\n" if $this->{id} == 2 ;
+          
+    { lock( @$jobs ) ;
+    
+      #print "WAIT...<<< $this->{id} [$last_job_is_standby] [". join('',@$jobs) ."] [${$this->{jobs_sz}}]\n" if $this->{id} == 2 ;
+      (!@$jobs_standby || $last_job_is_standby) && ${$this->{jobs_sz}} < 1 ? ($is_mother_thread ? cond_wait( @$jobs ) : cond_timedwait( @$jobs , time + 1 )) : undef ;
+      #print "WAIT...>>> $this->{id}\n" if $this->{id} == 2 ;
+    
+      #print "RUN...<<< $this->{id}\n" if $this->{id} == 2 ;
+    
+      last if $$status <= 0 ;
 
-    !@$jobs ? ($is_mother_thread ? cond_wait( @$jobs ) : cond_timedwait( @$jobs , time + 2 )) : undef ;
+      ##print threads->self->tid . "> RUN> $this->{id} [$MOTHER_THREAD]\n" ;
+      
+      ## !join('',@$jobs)
+      
+      if ( !join('',@$jobs) && @$jobs_standby ) {
+        ++$jobs_standby_i ;
+        $jobs_standby_i = $#{$jobs_standby} if $jobs_standby_i > $#{$jobs_standby} ;
+        $standby_job = $$jobs_standby[$jobs_standby_i] ;
+        
+        if ( $standby_job ) {
+          my $last_time ;
+          if ( $$standby_job[7] =~ /^(\d+)(?::(\d+))?$/s ) {
+            ( $standby_job_delay , $last_time ) = ( $1 , $2 ) ;
+            $standby_job = undef if (time - $last_time) < $standby_job_delay ;
+          }
+        }
+      }
+      
+      if ( !$standby_job ) {
+        next if ${$this->{jobs_sz}} < 1 ;
+        
+        ## Fix memory leak on Perl-5.8.4 Win32. (shift on shared array only OK for 5.8.6)
+        ##my $the_job = shift(@$jobs) ;
+        $the_job = $this->_jobs_shift ;
+        
+        next if !defined $the_job ;
+      }
+      
+    }
     
-    last if $$status <= 0 ;
-
-    ##print threads->self->tid . "> RUN> $this->{id} [$MOTHER_THREAD]\n" ;
+    if ( $standby_job ) {
+      $this->process_job($standby_job , \$running , 1) ;
+      $$standby_job[7] = "$standby_job_delay:" . time() if $standby_job_delay ;
+      $last_job_is_standby = 1 ;
+    }
+    elsif ( $the_job ) {
+      ##print $the_job->dump ;
+      $this->process_job($the_job , \$running) ;
+      $last_job_is_standby = undef ;
+    }
     
-    next if !@$jobs ;
-    
-    ## Fix memory leak on Perl-5.8.4 Win32. (shift on shared array only OK for 5.8.6)
-    ##my $the_job = shift(@$jobs) ;
-    my $the_job = $this->_jobs_shift ;
-  
-    next if !defined $the_job ;
-    
-    ##print $the_job->dump ;
-    
-    $this->process_job($the_job , \$running) ;
   }
   
   $$status = 0 ;
@@ -751,10 +1025,13 @@ sub process_job {
   my $this = shift ;
   my $the_job = shift ;
   my $running = shift ;
+  my $is_standby = shift ;
 
   return if !defined $the_job ;
 
   my ($jobs , $job_now , $err) = @$this{qw(jobs job_now err)} ;
+    
+  ##print $the_job->dump ;
   
   $$the_job[2] = 1 ;
 
@@ -764,8 +1041,8 @@ sub process_job {
   my $job_type = $$the_job[3] ;
   my @args = Thread::Isolate::thaw( $$the_job[4] ) ;
   
-  ##print $the_job->dump ;
-  
+  my $hold_args = $is_standby ? $$the_job[4] : undef ;
+    
   { lock( $$err ) ;
     
     if ($job_type eq 'SHUTDOWN') {
@@ -802,9 +1079,13 @@ sub process_job {
     }
   }
   
-  { lock( @$the_job ) ;
+  {
+    lock( @$the_job ) ;
     $$the_job[2] = 2 ;
     $$job_now = undef ;
+    
+    $$the_job[4] = $hold_args if $hold_args ;
+    
     cond_signal( @$the_job ) ;
   }
   
